/*
								+---------------------------------+
								|                                 |
								|    ***   PADS-like menu  ***    |
								|                                 |
								|  Copyright   -tHE SWINe- 2006  |
								|                                 |
								|             Menu.h              |
								|                                 |
								+---------------------------------+
*/

#ifndef __MENU_INCLUDED
#define __MENU_INCLUDED

/*
 *	char *p_s_Clone(const char *p_s_str)
 *		- alloc a new bufer and copy the contents
 */
char *p_s_Clone(const char *p_s_str);

enum {
	watch_Int,
	watch_Float,
	watch_String
};

enum {
	bf_Static = 0,
	bf_Callback = 1,
	bf_Link = 2,
	bf_ShowWatch = 4,
	bf_Sel_ToggleVal = 16, // int
	bf_Sel_RangeVal = 32, // float
	bf_Sel_EditStr = 64, // char*
	bf_Sel_Mask = 64 + 16 + 32
};
// button-flag

struct TToggleChoice {
	char *p_s_name;
	int n_value;

	TToggleChoice();
	bool Load(/*const*/ CTextParser &r_parser);
	void Free();
};
// single choice of toggle-val button

struct TToggleGroup {
	int n_id;
	std::vector<TToggleChoice> choice_list;

	TToggleGroup();
	bool Load(/*const*/ CTextParser &r_parser);
	const char *p_s_Translate(int n_value) const;
	int n_FirstValue(int n_value) const;
	int n_LastValue(int n_value) const;
	int n_PrevValue(int n_value) const;
	int n_NextValue(int n_value) const;
	int n_Value_Num() const;
	int n_Value_Index(int n_value) const;
	int n_Value(int n_index) const;
	void Free();
};
// group of toggle choices, identified by it's id

struct TWatchInfo {
	int n_data_type;
	char *p_s_name;
	union {
		int *p_n_value;
		float *p_f_value;
		std::string *p_s_value;
	};

	TWatchInfo();
	bool b_Initialized() const;
	bool Load(/*const*/ CTextParser &r_parser);
	void Free();
};
// watches are registered using this structure

enum {
	event_ButtonActivate,
	event_EditBegin,
	event_EditChange,
	event_EditEnd
};

struct TMenuButton;

struct TCallbackHandler {
	virtual bool b_Handle(const char *p_s_function_name) const = 0; // return true if it wants to get messages from this function
	virtual bool operator ()(const char *p_s_function_name, const TMenuButton *p_source, int n_event) = 0; // return true if it's allright to do the action (to jump trough menu, to start editing, to end editing ... ignored on event_EditChange)
};
// virtual class, will be used for connecting callback receiving class with menu

struct TMenuPage;

class CPadsMenu {
public:
	struct TMenuMetrics {
		float f_width;
		float f_title_height;
		float f_button_height;
		Vector2f v_button_spacing;
		Vector2f v_menu_offset;

		bool Load(/*const*/ CTextParser &r_parser);
	};

	struct TMenuColor {
		Vector3f v_bk_color;
		Vector3f v_focus_color;
		Vector3f v_edit_color;
		Vector3f v_border_color;
		Vector3f v_font_color;

		float f_alpha;

		bool Load(/*const*/ CTextParser &r_parser);
	};

protected:
	std::vector<TToggleGroup> m_enum_list;

	std::vector<TMenuPage*> m_page_list;
	TMenuPage *m_p_cur_page;
	int m_n_first_page_id;

	TMenuMetrics m_t_metrics;
	TMenuColor m_t_color;

	CStrokeFont *m_p_stroke_font;

	int m_n_button_num;

public:
	CPadsMenu();
	~CPadsMenu();

	bool Load(/*const*/ CTextParser &r_parser);
	bool Load(const char *p_s_filename);
	bool Navigate_FirstPage();
	void Draw(CGLState *p_state);
	void Free();

	void OnMouseMove(Vector2f v_pos);
	void OnMouseDown(Vector2f v_pos);
	void OnMouseUp(Vector2f v_pos);
	void OnKeyPress(int n_code);

	int n_CurButton_Token() const;
	// return flag of either button being editted or button being
	// highlighted in case no button is edited

	int n_RegisterCallback(TCallbackHandler *p_callback_handler); // returns number of menu buttons that will use this callback
	void RemoveCallback(TCallbackHandler *p_callback_handler);

	int n_AddWatch(const char *p_s_name, int *p_value_ptr);
	int n_AddWatch(const char *p_s_name, float *p_value_ptr);
	int n_AddWatch(const char *p_s_name, std::string *p_value_ptr); // returns number of menu buttons, referencing this watch
	void RemoveWatch(const char *p_s_name);
	// if you want to change address of watch, you have to call n_AddWatch again,
	// you have to remember the value of the old watch, if you want to keep it
};

class CSlider;
class CEditBox;

struct TMenuButton {
protected:
	char *p_s_name;
	char *p_s_name_buffer; // buffer for displayed name in case watch is used
	int n_name_length; // length of buffer, p_s_name is allocated for
	int n_flags;
	int n_preferred_index;
	int n_token;

	struct TLinkInfo {
		int n_target;
	} t_link;

	struct TCallbackInfo {
		char *p_s_name;
		TCallbackHandler *p_handler;

		TCallbackInfo()
			:p_s_name(0)
		{
		}
	} t_callback;

	TWatchInfo t_watch;

	struct TSelInfo {
		struct TToggleInfo {
			int n_group_id;
			const TToggleGroup *p_toggle_group;
		} t_toggle;

		struct TRangeInfo {
			float f_min, f_max, f_step;
			CSlider *p_slider;

			TRangeInfo()
				:p_slider(0)
			{
			}
		} t_range;

		struct TEditStringInfo {
			int n_max_length; // watch allocated by user must have max length set
			
			CEditBox *p_edit_box;
		} t_edit_str;

		int n_type;
		bool b_active;

		TWatchInfo t_sel_value;

		TSelInfo()
			:n_type(0), b_active(false)
		{
		}
	} t_sel;

public:
	TMenuButton();
	bool Load(/*const*/ CTextParser &r_parser);
	void Draw(CGLState *p_state, const CStrokeFont *p_font, Vector2f v_min,
		Vector2f v_max, const CPadsMenu::TMenuColor &r_t_menu_color, bool b_highlight);
	void Free();

	bool b_IsEdited() const;

	int n_OnMouseMove(Vector2f v_pos, Vector2f v_min, Vector2f v_max);
	int n_OnMouseDown(Vector2f v_pos, Vector2f v_min, Vector2f v_max);
	int n_OnMouseUp(Vector2f v_pos, Vector2f v_min, Vector2f v_max); // sent when being edited or active button
	int n_OnKeyPress(int n_code, char n_char); // sent only if it's button being edited
	// returns -1 / menu page where it wantrs to jump

	int n_Token() const { return n_token; }
	int n_PreferredIndex() const { return n_preferred_index; }
	void SetIndex(int n_index) { n_preferred_index = n_index; }

	bool Find_ToggleGroup(const std::vector<TToggleGroup> &r_enum_list);

	int n_RegisterCallback(TCallbackHandler *p_callback_handler); // returns number (0 or 1 here) of menu buttons that will use this callback
	void RemoveCallback(TCallbackHandler *p_callback_handler);

	int n_AddWatch(const char *p_s_name, int *p_value_ptr);
	int n_AddWatch(const char *p_s_name, float *p_value_ptr);
	int n_AddWatch(const char *p_s_name, std::string *p_value_ptr); // returns number (0 or 1 here) of menu buttons, referencing this watch
	void RemoveWatch(const char *p_s_name);

protected:
	bool DoCallback(int n_event);
};
// menu button

struct TMenuPage {
protected:
	int n_id;
	char *p_s_name;
	std::vector<TMenuButton> button_list;
	int n_active_button;
	TMenuButton *p_edit_button;

public:
	TMenuPage();
	bool Load(/*const*/ CTextParser &r_parser);
	void Draw(CGLState *p_state, const CStrokeFont *p_font,
		const CPadsMenu::TMenuMetrics &r_t_menu_metrics,
		const CPadsMenu::TMenuColor &r_t_menu_color, int n_button_num);
	void Free();

	int n_OnMouseMove(Vector2f v_pos, const CPadsMenu::TMenuMetrics &r_t_menu_metrics, int n_button_num);
	int n_OnMouseDown(Vector2f v_pos, const CPadsMenu::TMenuMetrics &r_t_menu_metrics, int n_button_num);
	int n_OnMouseUp(Vector2f v_pos, const CPadsMenu::TMenuMetrics &r_t_menu_metrics, int n_button_num);
	int n_OnKeyPress(int n_code, int n_button_num);
	// returns -1 / menu page where it wantrs to jump

	Vector2f v_ButtonMin(const CPadsMenu::TMenuMetrics &r_t_menu_metrics,
		int n_button_num, int n_button_index) const;
	Vector2f v_ButtonMax(const CPadsMenu::TMenuMetrics &r_t_menu_metrics,
		int n_button_num, int n_button_index) const;
	int n_ButtonIndex(const CPadsMenu::TMenuMetrics &r_t_menu_metrics,
		int n_button_num, Vector2f v_mouse) const;

	int n_Id() const { return n_id; }
	int n_CurButton_Token() const { if(n_active_button != -1) return button_list[n_active_button].n_Token(); return 0; }

	bool Find_ToggleGroups(const std::vector<TToggleGroup> &r_enum_list);

	int n_RegisterCallback(TCallbackHandler *p_callback_handler); // returns number of menu buttons that will use this callback
	void RemoveCallback(TCallbackHandler *p_callback_handler);

	int n_AddWatch(const char *p_s_name, int *p_value_ptr);
	int n_AddWatch(const char *p_s_name, float *p_value_ptr);
	int n_AddWatch(const char *p_s_name, std::string *p_value_ptr); // returns number of menu buttons, referencing this watch
	void RemoveWatch(const char *p_s_name);

protected:
	TMenuButton *p_Button(int n_index);
};

class CSlider {
protected:
	float *m_p_ref_value, m_f_min, m_f_max, m_f_step;
	float m_f_snap;
	bool m_b_snapped;

public:
	CSlider(float *p_ref_value, float f_min, float f_max, float f_step, float f_snap = .05f);

	void Draw(CGLState *p_state, Vector2f v_min,
		Vector2f v_max, const CPadsMenu::TMenuColor &r_t_menu_color);

	void OnMouseMove(Vector2f v_pos, Vector2f v_min, Vector2f v_max);
	void OnMouseDown(Vector2f v_pos, Vector2f v_min, Vector2f v_max);
	void OnMouseUp(Vector2f v_pos, Vector2f v_min, Vector2f v_max);
	void OnKeyPress(int n_code, char n_char);
};

class CCharFilter {
public:
	virtual bool b_Accept(char c_char, const char *p_s_string, int n_insert_pos) = 0;
	// return true if char can be inserted to index n_insert_pos (other chars are shifted)

	virtual bool b_Accept(char c_char, const char *p_s_string,
		int n_exclude_begin, int n_exclude_num) = 0;
	// return true if char can be inserted to index n_exclude_begin (other chars are shifted)
	// while n_exclude_num characters will be erased by insertion
};

class CEditBox {
protected:
	char *m_p_s_text;
	int m_n_alloc;
	int m_n_max_length;

	int m_n_cursor_pos; // !! cursor can be on both sides of selection
	int m_n_sel_start;
	int m_n_sel_length;

	CCharFilter *m_p_filter;

	float m_f_last_pos;

	bool m_b_dirty;

public:
	CEditBox(const char *p_s_initial_text, CCharFilter *p_filter = 0, int n_max_length = -1);
	~CEditBox();

	void Draw(CGLState *p_state, const CStrokeFont *p_font, Vector2f v_min,
		Vector2f v_max, const CPadsMenu::TMenuColor &r_t_menu_color);

	bool b_Dirty();
	void ResetDirtyFlag();
	const char *p_s_Text();

	void Delete_Selection();

	void OnMouseMove(Vector2f v_pos, Vector2f v_min, Vector2f v_max);
	void OnMouseDown(Vector2f v_pos, Vector2f v_min, Vector2f v_max);
	void OnMouseUp(Vector2f v_pos, Vector2f v_min, Vector2f v_max);
	void OnKeyPress(int n_code, char n_char);
};

#endif //__MENU_INCLUDED
