/*
								+----------------------------------+
								|                                  |
								| ***  Pashik virtual machine  *** |
								|                                  |
								|   Copyright  -tHE SWINe- 2005   |
								|                                  |
								|          TextParser2.h           |
								|                                  |
								+----------------------------------+
*/

#include "Pashik.h"

#ifndef __TEXT_PARSER_2_INCLUDED
#define __TEXT_PARSER_2_INCLUDED

enum {
    key_node,
    key_char,
    key_int,
    key_float,
    key_string,
    key_int_array, // nejsou pro pole typu name[], ale name = {x, y, z, bla};
    key_float_array,
    key_string_array
};
// datov typy

struct TNode;

struct TKey {
    int n_line;
    char *p_s_name;
    TNode *p_parent;
    //
    int n_type;
    union {
        int n_value;
        float f_value;
        char *p_s_value;
        std::vector<int> *p_int_list;
        std::vector<float> *p_float_list;
        std::vector<char*> *p_string_list;
    } t_value;

    virtual int b_IsNode() { return false; }
    virtual int b_IsKeyArray() { return false; }
    virtual int n_Index() { return -1; }
    void Free(); // volatel mus smazat po zavoln jet this!
};
// obecn kl - dka, jmno a rodi + typ a hodnota

struct TNode : TKey {
    std::vector<TKey*> m_key_list;

    virtual int b_IsNode() { return true; }
    void Free();
};
// noda - m seznam kl

enum {
	Parser_Ok = Pash_Ok,
	Parser_NoMem = Pash_NoMem,		// nen pam
	Parser_NoFile = Pash_NoFile,	// soubor neexistuje
	Parser_Syntax = Pash_Syntax,	// syntaktick chyba
	Parser_NoKey = Pash_Parser,		// kl nenalezen
	Parser_Error = Pash_Error,		// generick chyba
	Parser_UnexpectedToken = Pash_NumErrors, // neoekvan token
	Parser_MissingEndStmt,			// chyb stednk
	Parser_UnexpectedEOF,			// neoekvan konec souboru
	Parser_NegativeIndex,			// zporn index u pole
//	Parser_EmptyArray,				// pole nesm bt przdn ("blah = {};") - u nen chyba, konvertuje na intov pole
	Parser_BadArrayTypes,			// v poli jsou najednou floaty a stringy
	Parser_NoNullStringByInt,		// v poli string je nenulov int konstanta (nezastav parsovn, ale nahrad konstantu nulou)
	Parser_NoConversion,			// poadovan poloka m jin typ, ne pedpokldan
	Parser_OutOfSpace				// pi ukldn na disk dolo msto
};

class CSymbolTable;
class CTextParser {
protected:
	TNode m_t_root;
	TKey m_t_temp; // odkazy-vracejc funkce vracej hodnoty tohoto kle, pokud se kl nenael
	//
	TKey *m_p_locked_key;
	//
	char *m_p_s_ini_file;
	//
	int m_n_last_error;
	int m_n_error_line;
	//
	std::vector<Token> m_token_list;
	int m_n_cur_token;
	CSymbolTable *m_p_enum_table;
	//
	static const int n_min_records_to_open_branch; // minimln poet poloek nody pro pouit vtve msto tekov notace (pi zapisovn do souboru)

public:
	CTextParser();
	~CTextParser();

	int Attach(const char *p_s_ini_file);
	int Parse(const char *p_s_text, int n_length); // jen pro ten, nejde ukldat (detach vrac false, ale data uvoln)
	int Detach();

	int LoadFromFile();
	int SaveToFile(const char *p_s_comment_msg = 0, int n_tab_spaces = 4); // n_tab_spaces : 4 = default, -1 = insert tabs

	const char *p_s_Filename() { return m_p_s_ini_file; }

	int n_GetLastError(); // vrt posledn chybu
	int n_GetLastError(int &r_n_last_error_line); // vrt posledn chybu spolu s dkou kde se stala

	int b_IsIdent(const char *p_s_key); // true, pokud je jmno identifiktor a je k nmu hodnota
	int b_IsNode(const char *p_s_key); // true, pokud je jmno koen ("name.*" nebo "name {*};")
	int LockNode(const char *p_s_key); // te jen z koenu, danho jmnem
	int LockNextNode(); // uzamkne dal nodu se stejnm jmnem; vrac false pokud u takov noda nen, nebo pokud nebyla dn zamkl
	int LockParentNode(); // uzamkne nodu o jedno v; vrac false, pokud nen noda zamknut; pokud se dostane a k root, odemkne
	int UnlockNode(); // odemkne posledn zamknut node a pejde o level v

	int b_IsArrayOfKeys(const char *p_s_key); // vrt true, pokud se vyskytuje name[0-9]
	int n_KeyArrayLen(const char *p_s_key); // vrt poet prvk pole, zan hledat u nuly a kon na mezee

	int n_Value(const char *p_s_key, int n_default_value, int b_add_new = true);
	float f_Value(const char *p_s_key, float f_default_value, int b_add_new = true);
	int &r_n_Value(const char *p_s_key, int n_default_value, int b_add_new = true);
	float &r_f_Value(const char *p_s_key, float f_default_value, int b_add_new = true);
	const char *p_s_Value(const char *p_s_key, const char *p_s_default_value, int b_add_new = true);
	int Set_StrValue(const char *p_s_key, const char *p_s_value, int b_add_new = true);
	int b_IsArray(const char *p_s_key);
	int n_ArrayLen(const char *p_s_key);
	const std::vector<int> *p_n_Array(const char *p_s_key, int n_default_size, const int *p_default_data, int b_add_new = true);
	const std::vector<float> *p_f_Array(const char *p_s_key, int n_default_size, const float *p_default_data, int b_add_new = true);
	const std::vector<char*> *p_s_Array(const char *p_s_key, int n_default_size, const char **p_default_data, int b_add_new = true);

	int DeleteKey(const char *p_s_key); // smae kl
	int DeleteKeyArray(const char *p_s_key); // smae pole kl. p_s_key mus konit "[0]"

protected:
	TKey *p_add_key(const char *p_s_key);
	TKey *p_add_key_r(TKey *p_searched_key, TKey *p_cur_key);

	TKey *p_find_key(const char *p_s_key);
	TKey *p_find_key_r(TKey *p_searched_key, TKey *p_cur_key);

	int Read_Tokens(const char *p_s_buffer, int n_length);
	int Read_Tokens(FILE *p_fr);

	void PushToken();
	Token t_PopToken();
	Token t_CurToken();

	TKey *CTextParser::p_Parse_r(TNode *p_node);

	int Parse_Tokens();

	int Write_KeyName(FILE *p_fw, TKey *p_key);
	int Write_String(FILE *p_fw, const char *p_s_string);
	int SaveToFile_r(FILE *p_fw, TKey *p_key, const char *p_s_spacer, int n_level);
};

struct TIniFile;

class CIniGuard {
protected:
	static std::vector<TIniFile*> m_file_list;
	static int m_n_instance_num;
	//
	TIniFile *m_p_cur_file;
	//
	static int m_n_default_int;
	static float m_f_default_float;

public:
	CIniGuard();
	~CIniGuard();

	int Attach(const char *p_s_ini_file);
	int Parse(char *p_s_text, int n_length); // jen pro ten, nejde ukldat (detach vrac false)
	void Detach();

	int LoadFromFile();
	int SaveToFile(char *p_s_comment_msg = 0, int n_tab_spaces = 4); // n_tab_spaces : 4 = default, -1 = insert tabs

	int n_GetLastParserError(); // vrt posledn chybu
	int n_GetLastParserError(int &r_n_last_error_line); // vrt posledn chybu spolu s dkou kde se stala

	int LockNode(const char *p_s_key); // te jen z koenu, danho jmnem
	int LockNextNode(); // uzamkne dal nodu se stejnm jmnem; vrac false pokud u takov noda nen, nebo pokud nebyla dn zamkl
	int LockParentNode(); // uzamkne nodu o jedno v; vrac false, pokud nen noda zamknut; pokud se dostane a k root, odemkne
	int UnlockNode(); // odemkne posledn zamknut node a pejde o level v

	int b_IsIdent(const char *p_s_key); // true, pokud je jmno identifiktor a je k nmu hodnota
	int b_IsNode(const char *p_s_key); // true, pokud je jmno koen ("name.*" nebo "name {*};")
	int b_IsArrayOfKeys(const char *p_s_key); // vrt true, pokud se vyskytuje name[0]; kl mus mt na konci [0]
	int n_KeyArrayLen(const char *p_s_key); // vrt poet prvk pole, zan hledat u nuly a kon na mezee; kl mus mt na konci [0]

	int n_Value(const char *p_s_key, int n_default_value, int b_add_new = true);
	float f_Value(const char *p_s_key, float f_default_value, int b_add_new = true);
	int &r_n_Value(const char *p_s_key, int n_default_value, int b_add_new = true);
	float &r_f_Value(const char *p_s_key, float f_default_value, int b_add_new = true);

	const char *p_s_Value(const char *p_s_key, const char *p_s_default_value, int b_add_new = true);
	int Set_StrValue(const char *p_s_key, const char *p_s_value, int b_add_new = true);

	int b_IsArray(const char *p_s_key);
	int n_ArrayLen(const char *p_s_key);
	const std::vector<int> *p_n_Array(const char *p_s_key, int n_default_size, const int *p_default_data, int b_add_new = true);
	const std::vector<float> *p_f_Array(const char *p_s_key, int n_default_size, const float *p_default_data, int b_add_new = true);
	const std::vector<char*> *p_s_Array(const char *p_s_key, int n_default_size, const char **p_default_data, int b_add_new = true);

	int DeleteKey(const char *p_s_key); // smae kl
	int DeleteKeyArray(const char *p_s_key); // smae pole kl. p_s_key mus konit "[0]"
};

#endif // __TEXT_PARSER_2_INCLUDED